"Dictionaries to be used in multiple files."

EMSOutputVariableZone_dict = {
    'Comfortable Hours_No Applicability': ['ComfHoursNoApp', 'H', 'Summed'],
    'Comfortable Hours_Applicability': ['ComfHours', 'H', 'Summed'],
    'Occupied Comfortable Hours_No Applicability': ['OccComfHoursNoApp', 'H', 'Summed'],
    'Occupied Discomfortable Hours_No Applicability': ['OccDiscomfHoursNoApp', 'H', 'Summed'],
    'Occupied Hours': ['OccHours', 'H', 'Summed'],
    'Discomfortable Applicable Hot Hours': ['DiscomfAppHotHours', 'H', 'Summed'],
    'Discomfortable Applicable Cold Hours': ['DiscomfAppColdHours', 'H', 'Summed'],
    'Discomfortable Non Applicable Hot Hours': ['DiscomfNonAppHotHours', 'H', 'Summed'],
    'Discomfortable Non Applicable Cold Hours': ['DiscomfNonAppColdHours', 'H', 'Summed'],
    'Zone Floor Area': ['ZoneFloorArea', 'm2', 'Averaged'],
    'Zone Air Volume': ['ZoneAirVolume', 'm3', 'Averaged'],
    'People Occupant Count': ['Occ_count', '', 'Summed'],
}
